/**********************************************************************
 *<
	FILE: ObjectDeform.h

	DESCRIPTION:	Includes for Plugins

	CREATED BY:

	HISTORY:

 *>	Copyright (c) 2000, All Rights Reserved.
 **********************************************************************/

#ifndef __OBJECTDEFORM__H
#define __OBJECTDEFORM__H

#include "Max.h"
#include "resource.h"
#include "istdplug.h"
#include "iparamb2.h"
#include "iparamm2.h"

#include "meshadj.h"
#include "XTCObject.h"
#include "modstack.h"


extern TCHAR *GetString(int id);

extern HINSTANCE hInstance;


#define OBJECTDEFORM_CLASS_ID	Class_ID(0x635ec495, 0x703d5bc)




#define PBLOCK_REF	0

#define MESH_CHUNK 0x1010

#define MESH_VERTEXCOUNT_CHUNK 0x1020
#define MESH_FACECOUNT_CHUNK 0x1030

#define MESH_VERTEXDATA_CHUNK 0x1040
#define MESH_FACEDATA_CHUNK 0x1050


class VNormal {
	public:
		Point3 norm;
		DWORD smooth;
		VNormal *next;
		BOOL init;

		VNormal() {smooth=0;next=NULL;init=FALSE;norm=Point3(0,0,0);}
		VNormal(Point3 &n,DWORD s) {next=NULL;init=TRUE;norm=n;smooth=s;}
		~VNormal() {delete next;}
		void AddNormal(Point3 &n,DWORD s);
		Point3 &GetNormal(DWORD s);
		void Normalize();
	};

class Leaf;

class BoundsList
{
public:
	int faceIndex;
	BoundsList()
		{
		faceIndex = -1;
		}
};

class Leaf
{

private:
	Box2D bounds[4];
	BOOL bottom;
	
public:

	Leaf *quad[4];
	Tab<BoundsList> faceIndex;

	Leaf(Point2 min, Point2 max)
		{
		bottom = FALSE;
		quad[0] = NULL;
		quad[1] = NULL;
		quad[2] = NULL;
		quad[3] = NULL;
		Point2 mid = (min+max)*0.5f;

		bounds[0].min = min;
		bounds[0].max = mid;

		bounds[1].min.x = mid.x;
		bounds[1].min.y = min.y;
		bounds[1].max.x = max.x;
		bounds[1].max.y = mid.y;

		bounds[2].min.x = min.x;
		bounds[2].min.y = mid.y;
		bounds[2].max.x = mid.x;
		bounds[2].max.y = max.y;

		bounds[3].min = mid;
		bounds[3].max = max;

		}
	BOOL InQuad(int i, Box2D box)
		{
		Box2D b = bounds[i];
		if (box.max.x < b.min.x) return FALSE;
		if (box.min.x > b.max.x) return FALSE;

		if (box.max.y < b.min.y) return FALSE;
		if (box.min.y > b.max.y) return FALSE;

		return TRUE;

		}
	int InWhichQuad(Point3 p)
		{
		if ( (p.x >=bounds[0].min.x) && (p.x < bounds[0].max.x) &&
			 (p.y >=bounds[0].min.y) && (p.y < bounds[0].max.y) )
			 return 0;
		else if ( (p.x >=bounds[1].min.x) && (p.x < bounds[1].max.x) &&
			 (p.y >=bounds[1].min.y) && (p.y < bounds[1].max.y) )
			 return 1;
		else if ( (p.x >=bounds[2].min.x) && (p.x < bounds[2].max.x) &&
			 (p.y >=bounds[2].min.y) && (p.y < bounds[2].max.y) )
			 return 2;
		else if ( (p.x >=bounds[3].min.x) && (p.x < bounds[3].max.x) &&
			 (p.y >=bounds[3].min.y) && (p.y < bounds[3].max.y) )
			 return 3;
		return -1;
		}

	Box2D GetBounds(int i) {	if ( (i>=0) && (i<4) ) 
								return bounds[i];
								else return Box2D();}
	
	Leaf *GetQuad(int i) {	if ( (i>=0) && (i<4) ) 
							return quad[i];
							else return NULL;
						}
	void SetQuad(int i, Leaf *l) { if ( (i>=0) && (i<4) ) quad[i] = l;}

	BOOL IsBottom() {return bottom;}
	void SetBottom (BOOL b) { bottom = b;}

};

class BoundsTree
{


	BOOL invertNormals;
	BOOL normalizeMesh;
	int depth;

	float uoffset, voffset, uscale, vscale;

	Tab<VNormal> vnorms;
	Tab<Point3> fnorms;
	

	

	Mesh msh;
	Mesh initMsh;

	void LayoutTree(Leaf *l, int depth);

	RecurseTree(Leaf *l, int index, Box2D box);
	AddFace(int index, Box2D box);
	Tab<Box2D> boundingBoxList;

	void BuildNormals();


public:


	Leaf *head;


	BoundsTree()
		{
		normalizeMesh = TRUE;
		uoffset = 0.0f;
		voffset = 0.0f;
		uscale = 1.0f;
		vscale = 1.0f;
		depth = 6;
		head = NULL;
		}
	DeleteTree(Leaf *l);
	UpdateTreeFromBaked();
	BuildTree(Mesh *m);

	SetInitMesh(Mesh *m);
	Mesh GetInitMesh();
	
	IOResult Load(ILoad *iload);
	IOResult Save(ISave *isave); 

	void CloneFrom(BoundsTree tree)
		{
		initMsh = tree.initMsh;
		}

	BOOL IsInsideMesh(Point3 hitPoint);
	BOOL HitMesh(Point3 hitPoint, float &z);
	BOOL HitMesh(Point3 hitPoint, Point3 &norm);


	void InvertNorms(BOOL n) { invertNormals = n;}

	void SetDepth(int d) { depth = d;}
	void SetNormalizeMesh(int n) { normalizeMesh = n;}
	void SetPlacement(float ou, float ov, float su, float sv) 
			{ 
			uoffset = ou;
			voffset = ov;
			uscale = su;
			vscale = sv;
			}

private:
};





class ObjectDeform;

class ObjectDeformValidatorClass : public PBValidator
{
public:
class ObjectDeform *mod;
private:
BOOL Validate(PB2Value &v) 
	{
	INode *node = (INode*) v.r;

	if (node->TestForLoop(FOREVER,(ReferenceMaker *) mod)!=REF_SUCCEED) return FALSE;

	return TRUE;

	};
};

class FrameList
{
public:
	int frame;
	Tab<int> index;
	Tab<Point3> points;
};



class ObjectDeform : public Modifier {
	private:

		Interval ivalid;
		int quadDepth;


		

	public:


		Tab<float> distList;
		BitArray processList;
		Tab<Point3> vecList;

		float standoff,falloff;
		HWND rolluphWnd;
		int startFrame, endFrame;
		float normDist;

		BOOL collisionType;
		BOOL dent;

		
		Tab<VNormal> baseVNorms;
		Tab<Point3> baseFNorms;	

		BoundsTree meshSearchTree;

		Matrix3 objtm;
		Matrix3 invobjtm;
		ObjectDeformValidatorClass validator;
		// Parameter block
		IParamBlock2	*pblock;	//ref 0


		static IObjParam *ip;			//Access to the interface
		
		// From Animatable
		TCHAR *GetObjectName() { return GetString(IDS_CLASS_NAME); }

		//From Modifier
		ChannelMask ChannelsUsed()  { return GEOM_CHANNEL|TOPO_CHANNEL; }
		//TODO: Add the channels that the modifier actually modifies
		ChannelMask ChannelsChanged() { return GEOM_CHANNEL; }
		//TODO: Return the ClassID of the object that the modifier can modify
		Class_ID InputType() {return triObjectClassID;}

		void ModifyObject(TimeValue t, ModContext &mc, ObjectState *os, INode *node);
		void NotifyInputChanged(Interval changeInt, PartID partID, RefMessage message, ModContext *mc);

		void NotifyPreCollapse(INode *node, IDerivedObject *derObj, int index);
		void NotifyPostCollapse(INode *node,Object *obj, IDerivedObject *derObj, int index);


		Interval LocalValidity(TimeValue t);

		// From BaseObject
		//TODO: Return true if the modifier changes topology
		BOOL ChangeTopology() {return FALSE;}		
		
		CreateMouseCallBack* GetCreateMouseCallBack() {return NULL;}

		BOOL HasUVW();
		void SetGenUVW(BOOL sw);


		void BeginEditParams(IObjParam *ip, ULONG flags,Animatable *prev);
		void EndEditParams(IObjParam *ip, ULONG flags,Animatable *next);


		Interval GetValidity(TimeValue t);

		// Automatic texture support
		
		// Loading/Saving
		IOResult Load(ILoad *iload);
		IOResult Save(ISave *isave);

		//From Animatable
		Class_ID ClassID() {return OBJECTDEFORM_CLASS_ID;}		
		SClass_ID SuperClassID() { return OSM_CLASS_ID; }
		void GetClassName(TSTR& s) {s = GetString(IDS_CLASS_NAME);}

		RefTargetHandle Clone( RemapDir &remap );
		RefResult NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
			PartID& partID,  RefMessage message);


		int NumSubs() { return 1; }
		TSTR SubAnimName(int i) { return GetString(IDS_PARAMS); }				
		Animatable* SubAnim(int i) { return pblock; }

		// TODO: Maintain the number or references here
		int NumRefs() { return 1; }
		RefTargetHandle GetReference(int i) { return pblock; }
		void SetReference(int i, RefTargetHandle rtarg) { pblock=(IParamBlock2*)rtarg; }




		int	NumParamBlocks() { return 1; }					// return number of ParamBlocks in this instance
		IParamBlock2* GetParamBlock(int i) { return pblock; } // return i'th ParamBlock
		IParamBlock2* GetParamBlockByID(BlockID id) { return (pblock->ID() == id) ? pblock : NULL; } // return id'd ParamBlock

		void DeleteThis() { delete this; }		
		//Constructor/Destructor

		ObjectDeform();
		~ObjectDeform();
		

		INode* GetNodeFromModContext(ModContext *smd, int &which);
		void BuildNormals(Mesh *mesh);



		BOOL inSolve;
		BOOL fromCache;
		int quality;
		Tab<FrameList*> frameData;
		int currentFrame;
		Tab<int> redirList;
		void SolveFrames();

		Tab<int> tempIndex;
		Tab<Point3> tempPoints;

		void DisplayCacheMem();
		void ClearCache();
		int CacheSize();



};


class ObjectDeformClassDesc:public ClassDesc2 {
	public:
	int 			IsPublic() { return TRUE; }
	void *			Create(BOOL loading = FALSE) { return new ObjectDeform(); }
	const TCHAR *	ClassName() { return GetString(IDS_CLASS_NAME); }
	SClass_ID		SuperClassID() { return OSM_CLASS_ID; }
	Class_ID		ClassID() { return OBJECTDEFORM_CLASS_ID; }
	const TCHAR* 	Category() { return GetString(IDS_CATEGORY); }

	const TCHAR*	InternalName() { return _T("ObjectDeform"); }	// returns fixed parsable name (scripter-visible name)
	HINSTANCE		HInstance() { return hInstance; }				// returns owning module handle

};


class MyEnumProc : public DependentEnumProc 
	{
      public :
      virtual int proc(ReferenceMaker *rmaker); 
      INodeTab Nodes;              
	  BOOL nukeME;
	  BOOL xRefFound;
	  int count;
	};

	

class ObjDeformer : public Deformer {
	public:		
		ObjectDeform *mod;
		ObjDeformer(ObjectDeform *m);
		Point3 Map(int i, Point3 p);
	};


//dialog stuff to get the Set Ref button
class ObjectDeformDlgProc : public ParamMap2UserDlgProc {
//public ParamMapUserDlgProc {
	public:
		ObjectDeform *mod;		
		ObjectDeformDlgProc(ObjectDeform *m) {mod = m;}		
		BOOL DlgProc(TimeValue t,IParamMap2 *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);		
		void DeleteThis() {delete this;}

	};



#endif // __OBJECTDEFORM__H
